@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2021 Red Hat, Inc.
+# Copyright (C) 2022 Red Hat, Inc.
 # Copyright (C) 2019 William Brown <william@blackhats.net.au>
 # All rights reserved.
 #
@@ -20,7 +20,8 @@ from lib389.monitor import MonitorLDBM
 from lib389.replica import Replicas
 from lib389.utils import ensure_str, is_a_dn, is_dn_parent
 from lib389.tasks import DBCompactTask
-from lib389._constants import *
+from lib389._constants import INSTALL_LATEST_CONFIG
+from lib389.properties import BACKEND_SAMPLE_ENTRIES
 from lib389.cli_base import (
     _format_status,
     _generic_get,
@@ -106,7 +107,7 @@ def _get_backend(inst, be_name):
         if (is_a_dn(be_name) and str2dn(be_suffix) == str2dn(be_name)) or (not is_a_dn(be_name) and cn == be_name):
             return be
 
-    raise ValueError('Could not find backend suffix: {}'.format(name))
+    raise ValueError('Could not find backend suffix: {}'.format(be_name))
 
 
 def _get_index(inst, be_name, attr):
@@ -235,14 +236,17 @@ def backend_delete(inst, basedn, log, args, warn=True):
     dn = _search_backend_dn(inst, args.be_name)
     if dn is None:
         raise ValueError("Unable to find a backend with the name: ({})".format(args.be_name))
-    if warn and args.json is False:
-        _warn(dn, msg="Deleting %s %s" % (SINGULAR.__name__, dn))
+    if not args.ack:
+        log.info("""Not removing backend: if you are really sure add: --do-it""")
+    else:
+        if warn and args.json is False:
+            _warn(dn, msg="Deleting %s %s" % (SINGULAR.__name__, dn))
 
-    be = _get_backend(inst, args.be_name)
-    _recursively_del_backends(be)
-    be.delete()
+        be = _get_backend(inst, args.be_name)
+        _recursively_del_backends(be)
+        be.delete()
 
-    log.info("The database, and any sub-suffixes, were sucessfully deleted")
+        log.info("The database, and any sub-suffixes, were successfully deleted")
 
 
 def backend_import(inst, basedn, log, args):
@@ -299,7 +303,7 @@ def backend_export(inst, basedn, log, args):
 def is_db_link(inst, rdn):
     links = ChainingLinks(inst).list()
     for link in links:
-        cn = link.get_attr_val_utf8('cn').lower()
+        cn = link.get_attr_val_utf8_l('cn')
         if cn == rdn.lower():
             return True
     return False
@@ -707,7 +711,6 @@ def backend_list_vlv(inst, basedn, log, args):
                     log.info(' - dn: ' + idx.dn)
                     for k, v in list(entry.items()):
                         log.info(' - {}: {}'.format(ensure_str(k), ensure_str(v[0])))
-                    log.info()
 
     if args.json:
         log.info(json.dumps({"type": "list", "items": results}, indent=4))
@@ -718,7 +721,7 @@ def backend_get_vlv(inst, basedn, log, args):
     be = _get_backend(inst, args.be_name)
     vlvs = be.get_vlv_searches()
     for vlv in vlvs:
-        vlv_name = vlv.get_attr_val_utf8_l('cn').lower()
+        vlv_name = vlv.get_attr_val_utf8_l('cn')
         if vlv_name == args.name.lower():
             if args.json:
                 entry = vlv.get_attrs_vals_json(VLV_SEARCH_ATTRS)
@@ -1152,6 +1155,9 @@ def create_parser(subparsers):
     delete_parser = subcommands.add_parser('delete', help='Delete a backend database')
     delete_parser.set_defaults(func=backend_delete)
     delete_parser.add_argument('be_name', help='The backend name or suffix')
+    delete_parser.add_argument('--do-it', dest="ack",
+                               help="Remove backend and its subsuffixes",
+                               action='store_true', default=False)
 
     #######################################################
     # Get Suffix Tree (for use in web console)
