@@ -10,7 +10,7 @@
 from lib389.schema import Schema, Resolver
 from lib389.backend import Backends
 from lib389.migrate.openldap.config import olOverlayType
-from lib389.plugins import MemberOfPlugin, ReferentialIntegrityPlugin, AttributeUniquenessPlugins
+from lib389.plugins import MemberOfPlugin, ReferentialIntegrityPlugin, AttributeUniquenessPlugins, PassThroughAuthenticationPlugin, PAMPassThroughAuthPlugin, PAMPassThroughAuthConfigs
 import ldap
 import os
 from ldif import LDIFParser
@@ -101,6 +101,7 @@ class ImportTransformer(LDIFParser):
     def __init__(self, f_import, f_outport, exclude_attributes_set, exclude_objectclass_set):
         self.exclude_attributes_set = exclude_attributes_set
         # Already all lowered by the db ldif import
+
         self.exclude_objectclass_set = exclude_objectclass_set
         self.f_outport = f_outport
         self.writer = LDIFWriter(self.f_outport)
@@ -139,6 +140,21 @@ class ImportTransformer(LDIFParser):
                 # Not found, move on.
                 pass
 
+        # If userPassword is present AND it is a SASL map, then we migrate it to the nsSaslauthId
+        # type to prevent account password confusion.
+        try:
+            pw_a = amap['userpassword']
+            pw = entry[pw_a][0]
+            if pw.startswith(b'{SASL}'):
+                entry.pop(pw_a)
+                sasl_id = pw.replace(b'{SASL}', b'')
+                # Add the marker objectClass and the map attr
+                entry[oc_a] += [b'nsSaslauthAccount']
+                entry['nsSaslauthId'] = [sasl_id]
+        except:
+            # Not found, move on.
+            pass
+
         # Write it out
         self.writer.unparse(dn, entry)
 
@@ -219,20 +235,21 @@ class SchemaAttributeInconsistent(MigrationAction):
         log.info(f" * Schema Skip Inconsistent Attribute -> {self.attr.names[0]} ({self.attr.oid})")
 
     def display_post(self, log):
-        log.info(f" * [ ] - Review Schema Inconsistent Attribute -> {self.obj.names[0]} ({self.obj.oid})")
+        log.info(f" * [ ] - Review Schema Inconsistent Attribute -> {self.attr.names[0]} ({self.attr.oid})")
 
 class SchemaAttributeAmbiguous(MigrationAction):
-    def __init__(self, attr):
+    def __init__(self, attr, overlaps):
         self.attr = attr
+        self.overlaps = overlaps
 
     def __unicode__(self):
         return f"SchemaAttributeAmbiguous -> {self.attr.__unicode__()}"
 
     def display_plan(self, log):
-        log.info(f" * Schema Skip Abmiguous Attribute -> {self.attr.names[0]} ({self.attr.oid})")
+        log.info(f" * Schema Skip Abmiguous Attribute -> {self.attr.names[0]} ({self.attr.oid}) overlaps {self.overlaps}")
 
     def display_post(self, log):
-        log.info(f" * [ ] - Review Schema Ambiguous Attribute -> {self.obj.names[0]} ({self.obj.oid})")
+        log.info(f" * [ ] - Review Schema Ambiguous Attribute -> {self.attr.names[0]} ({self.attr.oid}) overlaps {self.overlaps}")
 
 class SchemaAttributeInvalid(MigrationAction):
     def __init__(self, attr, reason):
@@ -455,6 +472,72 @@ class PluginPwdPolicyAudit(MigrationAction):
         log.info(f" * [ ] - Enable Password Policy for new accounts in {self.suffix}. See `dsconf {self.inst_name} localpwp --help` ")
 
 
+class PluginPassThroughDisable(MigrationAction):
+    def __init__(self):
+        pass
+
+    def apply(self, inst):
+        pta = PassThroughAuthenticationPlugin(inst)
+        pta.disable()
+
+    def __unicode__(self):
+        return "PluginPassThroughDisable"
+
+    def display_plan(self, log):
+        log.info(f" * Plugin:PassThrough Disable")
+
+    def display_post(self, log):
+        pass
+
+
+class PluginPAMPassThroughEnable(MigrationAction):
+    def __init__(self):
+        pass
+
+    def apply(self, inst):
+        pta = PAMPassThroughAuthPlugin(inst)
+        pta.enable()
+
+    def __unicode__(self):
+        return "PAMPassThroughAuthPlugin"
+
+    def display_plan(self, log):
+        log.info(f" * Plugin:PamPassThrough Enable")
+
+    def display_post(self, log):
+        pass
+
+
+class PluginPAMPassThroughConfigure(MigrationAction):
+    def __init__(self, suffix):
+        self.suffix = suffix
+
+    def apply(self, inst):
+        ptas = PAMPassThroughAuthConfigs(inst)
+        # PAMPassThroughAuthConfig
+        pta = ptas.ensure_state(properties = {
+            'cn': 'saslauthd',
+            'pamService': 'saslauthd',
+            'pamIncludeSuffix': self.suffix,
+            'pamFilter': '(objectClass=nsSaslauthAccount)',
+            'pamIdAttr': 'nsSaslauthId',
+            'pamIDMapMethod': 'ENTRY',
+            'pamSecure': 'TRUE',
+            'pamFallback': 'FALSE',
+            'pamModuleIsThreadSafe': 'TRUE',
+            'pamMissingSuffix': 'ALLOW',
+        })
+
+    def __unicode__(self):
+        return "PluginPAMPassThroughConfigure"
+
+    def display_plan(self, log):
+        log.info(f" * Plugin:PamPassThrough Configure")
+
+    def display_post(self, log):
+        log.info(f" * [ ] - Review Plugin:PAMPassThrough Migrated SASLauthd Configuration is Correct")
+
+
 class PluginUnknownManual(MigrationAction):
     def __init__(self, overlay):
         self.overlay = overlay
@@ -539,7 +622,16 @@ class Migration(object):
             '1.3.6.1.4.1.42.2.27.8.1.12', # pwdFailureCountInt
             '1.3.6.1.4.1.42.2.27.8.1.13', # pwdMustChange
             '1.3.6.1.4.1.42.2.27.8.1.14', # pwdAllowUserChange
+            '1.3.6.1.4.1.4754.2.99.1', # pwdPolicyChecker
             '1.3.6.1.4.1.42.2.27.8.2.1', # pwdPolicy objectClass
+            # Openldap supplies some schema which conflicts to ours, skip them
+            'NetscapeLDAPattributeType:198', # memberUrl
+            'NetscapeLDAPobjectClass:33', # groupOfURLs
+
+            # Dynamic Directory Services can't be supported due to missing syntax oid below, so we
+            # exclude the "otherwise" supported attrs / ocs
+            'DynGroupAttr:1', # dgIdentity
+            'DynGroupOC:1', # dgIdentityAux
         ] + skip_schema_oids)
         self._schema_oid_unsupported = set([
             # RFC4517 othermailbox syntax is not supported on 389.
@@ -561,6 +653,9 @@ class Migration(object):
             '1.3.6.1.4.1.42.2.27.8.1.15', # pwdSafeModify
             '1.3.6.1.4.1.4754.1.99.1', # pwdCheckModule
             '1.3.6.1.4.1.42.2.27.8.1.30', # pwdMaxRecordedFailure
+            # OpenLDAP dynamic directory services defines an internal
+            # oid ( 1.3.6.1.4.1.4203.666.2.7 )for dynamic group authz, but has very little docs about this.
+            'DynGroupAttr:2', # dgAuthz
         ])
 
         self._skip_entry_attributes = set(
@@ -574,12 +669,18 @@ class Migration(object):
                 'pwdsafemodify',
                 'pwdcheckmodule',
                 'pwdmaxrecordedfailure',
+                # dds attributes we don't support
+                'dgidentity',
+                'dgauthz'
             ] +
             [x.lower() for x in skip_entry_attributes]
         )
-        self._skip_entry_objectclasses = set(
+        # These tend to be be from overlays we don't support
+        self._skip_entry_objectclasses = set([
+            'pwdpolicychecker',
             'pwdpolicy',
-        )
+            'dgidentityaux'
+        ])
 
         self._gen_migration_plan()
 
@@ -635,7 +736,7 @@ class Migration(object):
                 self.plan.append(SchemaClassUnsupported(obj))
                 continue
             if obj.oid.lower().startswith('x-'):
-                self.plan.append(SchemaClassInvalid(attr, "descr-oid format is ambiguous with x- prefix. You MUST use a numeric oid."))
+                self.plan.append(SchemaClassInvalid(obj, "descr-oid format is ambiguous with x- prefix. You MUST use a numeric oid."))
                 continue
             # For the attr, find if anything has a name overlap in any capacity.
             overlaps = [ ds_obj for ds_obj in schema_objects if ds_obj.oid == obj.oid]
@@ -701,6 +802,12 @@ class Migration(object):
                 self.plan.append(PluginPwdPolicyAudit(oldb.suffix, self.inst.serverid))
             else:
                 raise Exception("Unknown overlay type, this is a bug!")
+        # We can't programatically detect the use of sasl authd. We can however
+        # create a configuration that "isn't harmful" in the case it's not used,
+        # and that does work in the case that it is present.
+        self.plan.append(PluginPassThroughDisable())
+        self.plan.append(PluginPAMPassThroughEnable())
+        self.plan.append(PluginPAMPassThroughConfigure(oldb.suffix))
 
 
     def _gen_db_plan(self):
