@@ -23,16 +23,9 @@ from subprocess import check_output, run, PIPE
 from lib389.passwd import password_generate
 from lib389._mapped_object_lint import DSLint
 from lib389.lint import DSCERTLE0001, DSCERTLE0002
-from lib389.utils import ensure_str, format_cmd_list
+from lib389.utils import ensure_str, format_cmd_list, DSVersion
 import uuid
 
-# Setuptools ships with 'packaging' module, let's use it from there
-try:
-    from pkg_resources.extern.packaging.version import LegacyVersion
-# Fallback to a normal 'packaging' module in case 'setuptools' is stripped
-except:
-    from packaging.version import LegacyVersion
-
 KEYBITS = 4096
 CA_NAME = 'Self-Signed-CA'
 CERT_NAME = 'Server-Cert'
@@ -132,7 +125,6 @@ class NssSsl(DSLint):
         :type alt_names: [str, ]
         :returns: String of the subject DN.
         """
-
         if self.dirsrv and len(alt_names) > 0:
             return SELF_ISSUER.format(GIVENNAME=self.dirsrv.get_uuid(), HOSTNAME=alt_names[0])
         elif len(alt_names) > 0:
@@ -250,7 +242,7 @@ only.
             openssl_version = check_output(['/usr/bin/openssl', 'version']).decode('utf-8').strip()
         except subprocess.CalledProcessError as e:
             raise ValueError(e.output.decode('utf-8').rstrip())
-        rehash_available = LegacyVersion(openssl_version.split(' ')[1]) >= LegacyVersion('1.1.0')
+        rehash_available = DSVersion(openssl_version.split(' ')[1]) >= DSVersion('1.1.0')
 
         if rehash_available:
             cmd = ['/usr/bin/openssl', 'rehash', certdir]
@@ -461,6 +453,98 @@ only.
             cert_values.append(re.match(r'^(.+[^\s])[\s]+([^\s]+)$', line.rstrip()).groups())
         return cert_values
 
+    def _openssl_get_csr_subject(self, csr_dir, csr_name):
+        cmd = [
+            '/usr/bin/openssl',
+            'req',
+            '-subject',
+            '-noout',
+            '-in',
+            '%s/%s'% (csr_dir, csr_name),
+        ]
+        self.log.debug("cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Parse the subject string from openssl output
+        result = result.replace("subject=", "")
+        result = result.replace(" ", "").strip()
+        result = result.split(',')
+        result = result[slice(None, None, -1)]
+        result = ','.join([str(elem) for elem in result])
+        return result
+
+    def _csr_show(self, name):
+        csr_dir = self.dirsrv.get_cert_dir()
+        result = ""
+        # Display PEM contents of a CSR file
+        if name and csr_dir:
+            if os.path.exists(csr_dir + "/" + name + ".csr"):
+                cmd = [
+                    "/usr/bin/sed",
+                    "-n",
+                    '/BEGIN NEW/,/END NEW/p',
+                    csr_dir + "/" + name + ".csr"
+                ]
+                self.log.debug("cmd: %s", format_cmd_list(cmd))
+                try:
+                    result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+                except subprocess.CalledProcessError as e:
+                    raise ValueError(e.output.decode('utf-8').rstrip())
+
+        return result
+
+    def _csr_list(self, csr_dir=None):
+        csr_list = []
+        csr_dir = self.dirsrv.get_cert_dir()
+        # Search for .csr file extensions in instance config dir
+        cmd = [
+            '/usr/bin/find',
+            csr_dir,
+            '-type',
+            'f',
+            '-name',
+            '*.csr',
+            '-printf',
+            '%f\\n',
+        ]
+        self.log.debug("cmd: %s", format_cmd_list(cmd))
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Bail out if we cant find any .csr files
+        if len(result) == 0:
+            return []
+
+        # For each .csr file, get last modified time and subject DN
+        for csr_file in result.splitlines():
+            csr = []
+            # Get last modified time stamp
+            cmd = [
+                '/usr/bin/date',
+                '-r',
+                '%s/%s'% (csr_dir, csr_file),
+                '+%Y-%m-%d %H:%M:%S',
+            ]
+            try:
+                result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+            except subprocess.CalledProcessError as e:
+                raise ValueError(e.output.decode('utf-8').rstrip())
+
+            # Add csr modified timestamp
+            csr.append(result.strip())
+            # Use openssl to get the csr subject DN
+            csr.append(self._openssl_get_csr_subject(csr_dir, csr_file))
+            # Add csr name, without extension
+            csr.append(csr_file.rsplit('.', 1)[0])
+            csr_list.append(csr)
+
+        return csr_list
+
     def _rsa_cert_key_exists(self, cert_tuple):
         name = cert_tuple[0]
         cmd = [
@@ -478,7 +562,7 @@ only.
             raise ValueError(e.output.decode('utf-8').rstrip())
         lines = result.split('\n')[1:-1]
         for line in lines:
-            m = re.match('\<(?P<id>.*)\> (?P<type>\w+)\s+(?P<hash>\w+).*:(?P<name>.+)', line)
+            m = re.match(r'\<(?P<id>.*)\> (?P<type>\w+)\s+(?P<hash>\w+).*:(?P<name>.+)', line)
             if name == m.group('name'):
                 return True
         return False
@@ -587,18 +671,22 @@ only.
         self.log.debug("nss output: %s", result)
         return True
 
-    def create_rsa_key_and_csr(self, alt_names=[], subject=None):
+    def create_rsa_key_and_csr(self, alt_names=[], subject=None, name=None):
         """Create a new RSA key and the certificate signing request. This
         request can be submitted to a CA for signing. The returned certificate
         can be added with import_rsa_crt.
         """
-        csr_path = os.path.join(self._certdb, '%s.csr' % CERT_NAME)
+        if name is None:
+            csr_path = os.path.join(self._certdb, '%s.csr' % CERT_NAME)
+        else:
+            csr_path = os.path.join(self._certdb, '%s.csr' % name)
 
         if len(alt_names) == 0:
             alt_names = self.detect_alt_names(alt_names)
         if subject is None:
             subject = self.generate_cert_subject(alt_names)
 
+        self.log.debug(f"CSR name -> {name}")
         self.log.debug(f"CSR subject -> {subject}")
         self.log.debug(f"CSR alt_names -> {alt_names}")
 
@@ -856,6 +944,52 @@ only.
         crt_der_path = '%s/%s%s.der' % (self._certdb, USER_PREFIX, name)
         return {'ca': ca_path, 'key': key_path, 'crt': crt_path, 'crt_der_path': crt_der_path}
 
+    def list_keys(self, orphan=None):
+        key_list = []
+        cmd = [
+            '/usr/bin/certutil',
+            '-K',
+            '-d',
+            self._certdb,
+            '-f',
+            '%s/%s' % (self._certdb, PWD_TXT),
+        ]
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        # Ignore the first line of certutil output
+        for line in result.splitlines()[1:]:
+            # Normalise the output of certutil
+            line = re.sub(r"\<[^>]*\>","", line)
+            key = re.split(r'\s{2,}', line)
+            if orphan:
+                if 'orphan' in line:
+                    key_list.append(key)
+            else:
+                key_list.append(key)
+
+        return key_list
+
+    def del_key(self, keyid):
+        cmd = [
+            '/usr/bin/certutil',
+            '-F',
+            '-d',
+            self._certdb,
+            '-f',
+            '%s/%s' % (self._certdb, PWD_TXT),
+            '-k',
+            keyid,
+        ]
+        try:
+            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
+        except subprocess.CalledProcessError as e:
+            raise ValueError(e.output.decode('utf-8').rstrip())
+
+        return result
+
     # Certificate helper functions
     def del_cert(self,  nickname):
         """Delete this certificate
@@ -1039,9 +1173,6 @@ only.
         if not os.path.exists(input_cert):
             raise ValueError("The cert file ({}) does not exist".format(input_cert))
 
-        self._assert_not_chain(input_key)
-        self._assert_not_chain(input_cert)
-
         self.log.debug(f"Importing key and cert -> {input_key}, {input_cert}")
 
         p12_bundle = "%s/temp_server_key_cert.p12" % self._certdb
@@ -1091,3 +1222,78 @@ only.
             # Remove the p12
             if os.path.exists(p12_bundle):
                 os.remove(p12_bundle)
+
+    def add_ca_cert_bundle(self, cert_file, nicknames):
+        """
+        Add a PEM file that could be a bundle of CA certs
+
+        :param nicknames: list of names of each CA certificate
+        :param cert_file: path to certificate PEM file
+        :raises:
+        """
+
+        # Verify input_file exists
+        if not os.path.exists(cert_file):
+            raise ValueError("The certificate file ({}) does not exist".format(cert_file))
+
+        try:
+            with open(cert_file, 'r') as f:
+                ca_count = 0
+                ca_files_to_cleanup = []
+                writing_ca = False
+                try:
+                    for line in f:
+                        if line.startswith('-----BEGIN CERTIFICATE-----'):
+                            # create tmp cert file
+                            writing_ca = True
+                            ca_file_name = cert_file + "-" + str(ca_count)
+                            ca_file = open(ca_file_name, 'w')
+                            ca_files_to_cleanup.append(ca_file_name)
+                            ca_file.write(line)
+                        elif writing_ca:
+                            ca_file.write(line)
+                            if line.startswith('-----END CERTIFICATE-----'):
+                                writing_ca = False
+                                ca_file.close()
+
+                                # Generate CA certificate nickname
+                                names_len = len(nicknames)
+                                if names_len > ca_count:
+                                    if nicknames[ca_count].lower() == CERT_NAME.lower() or nicknames[ca_count].lower() == CA_NAME.lower():
+                                        raise ValueError(f"You may not import a CA with the nickname {CERT_NAME} or {CA_NAME}")
+                                    ca_cert_name = nicknames[ca_count]
+                                else:
+                                    # A name was not provided for this cert, create one
+                                    ca_cert_name = nicknames[names_len - 1] + str(ca_count)
+
+                                # Check if certificate nickname exists
+                                name_exists = False
+                                try:
+                                    self.get_cert_details(ca_cert_name)
+                                    name_exists = True
+                                except ValueError:
+                                    pass
+
+                                if name_exists:
+                                    # Not good, cleanup and raise error
+                                    try:
+                                        for tmp_file in ca_files_to_cleanup:
+                                            os.remove(tmp_file)
+                                    except IOError:
+                                        # failed to remove tmp cert
+                                        log.debug("Failed to remove tmp cert file")
+                                        pass
+                                    raise ValueError(f"Certificate already exists with the same name ({ca_cert_name})")
+
+                                # Add this CA certificate
+                                self.add_cert(ca_cert_name, ca_file_name, ca=True)
+                                ca_count += 1
+                                log.info(f"Successfully added CA certificate ({ca_cert_name})")
+                except IOError as e:
+                    # Some failure, remove all the tmp files
+                    ca_file.close()
+                    for tmp_file in ca_file_to_cleanup:
+                        os.remove(tmp_file)
+                    raise e
+        except EnvironmentError as e:
+            raise e
