@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2019 Red Hat, Inc.
+# Copyright (C) 2022 Red Hat, Inc.
 # All rights reserved.
 #
 # License: GPL (version 3 or any later version).
@@ -12,10 +12,12 @@ import copy
 import os.path
 from lib389 import tasks
 from lib389._mapped_object import DSLdapObjects, DSLdapObject
-from lib389.lint import DSRILE0001, DSRILE0002
+from lib389.lint import DSRILE0001, DSRILE0002, DSMOLE0001
 from lib389.utils import ensure_str, ensure_list_bytes
 from lib389.schema import Schema
-from lib389._constants import DN_PLUGIN
+from lib389._constants import (
+        DN_PLUGIN, DN_MBO_TASK, DN_AUTOMEMBER_REBUILD_TASK, DN_FIXUP_LINKED_ATTIBUTES,
+        DN_EUUID_TASK)
 from lib389.properties import (
         PLUGINS_OBJECTCLASS_VALUE, PLUGIN_PROPNAME_TO_ATTRNAME,
         PLUGINS_ENABLE_ON_VALUE, PLUGINS_ENABLE_OFF_VALUE, PLUGIN_ENABLE
@@ -773,6 +775,58 @@ class MemberOfPlugin(Plugin):
         self._create_objectclasses.extend(['extensibleObject'])
         self._must_attributes.extend(['memberOfGroupAttr', 'memberOfAttr'])
 
+    @classmethod
+    def lint_uid(cls):
+        return 'memberof'
+
+    def _lint_member_attr_indexes(self):
+        if self.status():
+            from lib389.backend import Backends
+            backends = Backends(self._instance).list()
+            attrs = self.get_attr_vals_utf8_l("memberofgroupattr")
+            container = self.get_attr_val_utf8_l("nsslapd-plugincontainerscope")
+            for backend in backends:
+                suffix = backend.get_attr_val_utf8_l('nsslapd-suffix')
+                if suffix == "cn=changelog":
+                    # Always skip retro changelog
+                    continue
+                if container is not None:
+                    # Check if this backend is in the scope
+                    if not container.endswith(suffix):
+                        # skip this backend that is not in the scope
+                        continue
+                indexes = backend.get_indexes()
+                for attr in attrs:
+                    report = copy.deepcopy(DSMOLE0001)
+                    try:
+                        index = indexes.get(attr)
+                        types = index.get_attr_vals_utf8_l("nsIndexType")
+                        valid = False
+                        if "eq" in types:
+                            valid = True
+
+                        if not valid:
+                            report['detail'] = report['detail'].replace('ATTR', attr)
+                            report['detail'] = report['detail'].replace('BACKEND', suffix)
+                            report['fix'] = report['fix'].replace('ATTR', attr)
+                            report['fix'] = report['fix'].replace('BACKEND', suffix)
+                            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
+                            report['items'].append(suffix)
+                            report['items'].append(attr)
+                            report['check'] = f'memberof:attr_indexes'
+                            yield report
+                    except:
+                        # No index at all, bad
+                        report['detail'] = report['detail'].replace('ATTR', attr)
+                        report['detail'] = report['detail'].replace('BACKEND', suffix)
+                        report['fix'] = report['fix'].replace('ATTR', attr)
+                        report['fix'] = report['fix'].replace('BACKEND', suffix)
+                        report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
+                        report['items'].append(suffix)
+                        report['items'].append(attr)
+                        report['check'] = f'memberof:attr_indexes'
+                        yield report
+
     def get_attr(self):
         """Get memberofattr attribute"""
 
@@ -953,7 +1007,10 @@ class MemberOfPlugin(Plugin):
         task_properties = {'basedn': basedn}
         if _filter is not None:
             task_properties['filter'] = _filter
-        task.create(properties=task_properties)
+        try:
+            task.create(properties=task_properties)
+        except ldap.NO_SUCH_OBJECT:
+            raise ValueError("The fixup task can not be run, the memberOf plugin is not fully enabled.")
 
         return task
 
@@ -1084,13 +1141,15 @@ class AutoMembershipPlugin(Plugin):
     def __init__(self, instance, dn="cn=Auto Membership Plugin,cn=plugins,cn=config"):
         super(AutoMembershipPlugin, self).__init__(instance, dn)
 
-    def fixup(self, basedn, _filter=None):
+    def fixup(self, basedn, _filter=None, cleanup=False):
         """Create an automember rebuild membership task
 
         :param basedn: Basedn to fix up
         :type basedn: str
         :param _filter: a filter for entries to fix up
         :type _filter: str
+        :param cleanup: cleanup old group memberships
+        :type cleanup: boolean
 
         :returns: an instance of Task(DSLdapObject)
         """
@@ -1099,10 +1158,24 @@ class AutoMembershipPlugin(Plugin):
         task_properties = {'basedn': basedn}
         if _filter is not None:
             task_properties['filter'] = _filter
+        if cleanup:
+            task_properties['cleanup'] = "yes"
+
         task.create(properties=task_properties)
 
         return task
 
+    def abort_fixup(self):
+        """Create an automember abort rebuild task
+
+        :returns: an instance of Task(DSLdapObject)
+        """
+
+        task = tasks.AutomemberAbortRebuildTask(self._instance)
+        task.create()
+
+        return task
+
 
 class AutoMembershipDefinition(DSLdapObject):
     """A single instance of Auto Membership Plugin config entry
@@ -1423,7 +1496,7 @@ class PAMPassThroughAuthPlugin(Plugin):
         super(PAMPassThroughAuthPlugin, self).__init__(instance, dn)
 
 
-class PAMPassThroughAuthConfig(Plugin):
+class PAMPassThroughAuthConfig(DSLdapObject):
     """A single instance of PAM Pass Through Auth config entry
 
     :param instance: An instance
@@ -1432,24 +1505,11 @@ class PAMPassThroughAuthConfig(Plugin):
     :type dn: str
     """
 
-    _plugin_properties = {
-        'cn' : 'USN',
-        'nsslapd-pluginEnabled': 'off',
-        'nsslapd-pluginPath': 'libpam-passthru-plugin',
-        'nsslapd-pluginInitfunc': 'pam_passthruauth_init',
-        'nsslapd-pluginType': 'betxnpreoperation',
-        'nsslapd-plugin-depends-on-type': 'database',
-        'nsslapd-pluginId': 'PAM',
-        'nsslapd-pluginVendor': '389 Project',
-        'nsslapd-pluginVersion': '1.3.7.0',
-        'nsslapd-pluginDescription': 'PAM Pass Through Auth plugin'
-    }
-
     def __init__(self, instance, dn=None):
         super(PAMPassThroughAuthConfig, self).__init__(instance, dn)
         self._rdn_attribute = 'cn'
         self._must_attributes = ['cn']
-        self._create_objectclasses = ['top', 'extensibleObject', 'nsslapdplugin', 'pamConfig']
+        self._create_objectclasses = ['top', 'extensibleObject', 'pamConfig']
         self._protected = False
 
 
@@ -1464,7 +1524,7 @@ class PAMPassThroughAuthConfigs(DSLdapObjects):
 
     def __init__(self, instance, basedn="cn=PAM Pass Through Auth,cn=plugins,cn=config"):
         super(PAMPassThroughAuthConfigs, self).__init__(instance)
-        self._objectclasses = ['top', 'extensibleObject', 'nsslapdplugin', 'pamConfig']
+        self._objectclasses = ['top', 'extensibleObject', 'pamConfig']
         self._filterattrs = ['cn']
         self._scope = ldap.SCOPE_ONELEVEL
         self._childobject = PAMPassThroughAuthConfig
@@ -2071,6 +2131,78 @@ class DNAPluginSharedConfigs(DSLdapObjects):
         return co.create(properties, self._basedn)
 
 
+class MemberOfFixupTasks(DSLdapObjects):
+    """A DSLdapObjects entity which represents memberOf fixup tasks
+
+    :param instance: An instance
+    :type instance: lib389.DirSrv
+    :param task_dn: dn for a specific task
+    :type basedn: str
+    """
+
+    def __init__(self, instance, task_dn=None):
+        super(MemberOfFixupTasks, self).__init__(instance)
+        self._objectclasses = ['top']
+        self._filterattrs = ['cn']
+        self._childobject = DSLdapObject
+        self._basedn = DN_MBO_TASK
+        self._scope = ldap.SCOPE_ONELEVEL
+
+
+class AutoMembershipFixupTasks(DSLdapObjects):
+    """A DSLdapObjects entity which represents automember fixup tasks
+
+    :param instance: An instance
+    :type instance: lib389.DirSrv
+    :param task_dn: dn for a specific task
+    :type basedn: str
+    """
+
+    def __init__(self, instance, task_dn=None):
+        super(AutoMembershipFixupTasks, self).__init__(instance)
+        self._objectclasses = ['top']
+        self._filterattrs = ['cn']
+        self._childobject = DSLdapObject
+        self._basedn = DN_AUTOMEMBER_REBUILD_TASK
+        self._scope = ldap.SCOPE_ONELEVEL
+
+
+class LinkedAttributesFixupTasks(DSLdapObjects):
+    """A DSLdapObjects entity which represents automember fixup tasks
+
+    :param instance: An instance
+    :type instance: lib389.DirSrv
+    :param task_dn: dn for a specific task
+    :type basedn: str
+    """
+
+    def __init__(self, instance, task_dn=None):
+        super(LinkedAttributesFixupTasks, self).__init__(instance)
+        self._objectclasses = ['top']
+        self._filterattrs = ['cn']
+        self._childobject = DSLdapObject
+        self._basedn = DN_FIXUP_LINKED_ATTIBUTES
+        self._scope = ldap.SCOPE_ONELEVEL
+
+
+class EntryUUIDFixupTasks(DSLdapObjects):
+    """A DSLdapObjects entity which represents automember fixup tasks
+
+    :param instance: An instance
+    :type instance: lib389.DirSrv
+    :param task_dn: dn for a specific task
+    :type basedn: str
+    """
+
+    def __init__(self, instance, task_dn=None):
+        super(EntryUUIDFixupTasks, self).__init__(instance)
+        self._objectclasses = ['top']
+        self._filterattrs = ['cn']
+        self._childobject = DSLdapObject
+        self._basedn = DN_EUUID_TASK
+        self._scope = ldap.SCOPE_ONELEVEL
+
+
 class Plugins(DSLdapObjects):
     """A DSLdapObjects entity which represents plugin entry
 
